from hashlib import sha256
from typing import List

ADDRESS_VERSION_BYTE_N3 = '35'


def _base_58_encode(num):
    """Converts a number into Base-58."""
    alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

    encoded: List[str] = []
    alpha_cnt: int = len(alphabet)

    if num < 0:
        return ''

    while num >= alpha_cnt:
        mod = num % alpha_cnt
        num //= alpha_cnt
        encoded.append(alphabet[mod])

    if num > 0:
        encoded.append(alphabet[num])

    return ''.join(encoded[::-1])


def _base_58_decode(ins: str) -> int:
    """Converts a Base-58 encoded integer, as string, back to a number."""
    alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

    multi: int = 1
    decoded: int = 0
    alpha_cnt: int = len(alphabet)

    for char in ins[::-1]:
        decoded += multi * alphabet.index(char)
        multi *= alpha_cnt

    return decoded


def _pad_address(hex_string):
    raw_byes = bytearray.fromhex(ADDRESS_VERSION_BYTE_N3 + hex_string)
    hashed_address = sha256(raw_byes)
    double_hashed_address = sha256(hashed_address.digest())
    double_hashed_address_hex = double_hashed_address.digest().hex()
    return ADDRESS_VERSION_BYTE_N3 + hex_string + double_hashed_address_hex[0:8]


def _reverse_hex(hex_string):
    raw_bytes = bytearray.fromhex(hex_string)
    raw_bytes.reverse()
    return raw_bytes.hex()


def encode_hash(hash_hex: str):
    """
    Converts a wallet hash to a wallet address.
    E.g. a7271ccc82bb311db369719d1868e6ac19ece31f -> NNpbB25aNMY2pD5JkWeAEbsNfp7C3GoZF1

    :param hash_hex: The wallet hash as an hex string.
    :return: Returns the corresponding N3 wallet address.
    :rtype: str
    """

    reversed_hex = _reverse_hex(hash_hex)
    padded_address = _pad_address(reversed_hex)
    return _base_58_encode(int(padded_address, 16))


def decode_address(address: str):
    """
    Converts a wallet address to a wallet hash.
    E.g. NNpbB25aNMY2pD5JkWeAEbsNfp7C3GoZF1 -> a7271ccc82bb311db369719d1868e6ac19ece31f

    :param address: The N3 wallet address.
    :return: Returns the corresponding wallet hash as hex.
    :rtype: str
    """

    raw_hash = _base_58_decode(address)
    hash_hex = hex(raw_hash)[4:-8]
    hash_bytes = bytearray.fromhex(hash_hex)
    hash_bytes.reverse()
    return hash_bytes.hex()



import json
from base64 import b64encode


def create_update_invocation_file(
        contract_name: str,
        nef_file_path: str,
        manifest_file_path: str,
        operation='update'
) -> str:
    """
    Create a Neo Express invocation file that can be used to update a smart contract.

    :param contract_name: The name of the smart contract to update.
    :param nef_file_path: The path to the smart contract NEF file.
    :param manifest_file_path: The path to the smart contract manifest file.
    :param operation: The name of the operation/method on the smart contract.
    :return: Returns the invocation script as JSON.
    :rtype: str
    """
    invocation_dict = {
        'contract': contract_name,
        'operation': operation,
    }

    args = []

    with open(nef_file_path, 'rb') as file:
        raw_bytes = file.read()
        args.append({
            'type': 'ByteArray',
            'value': raw_bytes.hex(),
        })

    with open(manifest_file_path, 'rb') as file:
        raw_bytes = file.read()
        args.append({
            'type': 'ByteArray',
            'value': raw_bytes.hex(),
        })

    invocation_dict['args'] = args

    return json.dumps([invocation_dict])



from neo_toolbox.data_transformation import decode_address


def create_owner_bytes(address):
    """
    Creates a small template that can be used in Boa3 smart contracts to identify owner of a contract.

    :param address: The address of the owner of the contract. E.g. NNpbB25aNMY2pD5JkWeAEbsNfp7C3GoZF1
    :return: Returns a template that can be copied to the smart contract.
    :rtype: str
    """

    hash_hex = decode_address(address)
    hash_bytes = bytearray.fromhex(hash_hex)
    hash_bytes.reverse()

    as_string = str(bytes(hash_bytes))

    return f"OWNER = UInt160({as_string})"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neo_toolbox']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'neo-toolbox',
    'version': '0.1.6',
    'description': 'Tools for working with the Neo N3 blockchain.',
    'long_description': '# Neo Toolbox\n\nSmall toolkit for interacting with the Neo N3 blockchain.\n\n### Available tools\n\n1. [Data Transformation](#data_transformation)\n2. [Create Owner Bytes](#create_owner_bytes)\n3. [Create Update Invocation File](#create_update_invocation_file)\n\n\n\n#### <a name="data_transformation"></a>Data Transformation\n\nLets you convert between wallet hashes and addresses.\n\n##### Example\n\n```python\nfrom neo_toolbox.data_transformation import encode_hash, decode_address\n\n# Encode a wallet hash\noriginal_hash_hex = \'a7271ccc82bb311db369719d1868e6ac19ece31f\'\naddress = encode_hash(original_hash_hex)\nprint(address)\n# Outputs: \'NNpbB25aNMY2pD5JkWeAEbsNfp7C3GoZF1\'\n\n# Decode a wallet address\noriginal_address = \'NNpbB25aNMY2pD5JkWeAEbsNfp7C3GoZF1\'\nhash_hex = decode_address(original_address)\nprint(hash_hex)\n# Outputs: \'a7271ccc82bb311db369719d1868e6ac19ece31f\'\n```\n\n\n#### <a name="create_owner_bytes"></a>Create Owner Bytes\n\nThis helper function lets you convert a N3 wallet address to an python variable that can be used\nin your smart contract to verify the owner of the contract. This is very useful to add admin\nfunctionality to your smart contract.\n\n##### Example\n\n```python\nfrom neo_toolbox.create_owner_bytes import create_owner_bytes\n\noriginal_address = \'NNpbB25aNMY2pD5JkWeAEbsNfp7C3GoZF1\'\ncode_snippet = create_owner_bytes(original_address)\nprint(code_snippet)\n# Outputs: \'OWNER = UInt160(b"\\x1f\\xe3\\xec\\x19\\xac\\xe6h\\x18\\x9dqi\\xb3\\x1d1\\xbb\\x82\\xcc\\x1c\'\\xa7")\'\n```\n\n#### <a name="create_update_invocation_file"></a>Create Update Invocation File\n\nThis helper function lets you create a [Neo Express](https://github.com/neo-project/neo-express)\ninvocation file that can be used to update a smart contract.\n\n##### Example\n\n```python\nfrom neo_toolbox.create_update_invocation_file import create_update_invocation_file\n\nnef_file = \'path/to/sample_contract.nef\'\nmanifest_file = \'path/to/sample_contract.manifest.json\'\ntemplate = create_update_invocation_file(\'your_contract_name\', nef_file, manifest_file, operation=\'my_custom_update\')\nprint(template)\n# Outputs a JSON string: \'[{"contract": ...\'\n```\n\n\n*Created and maintained by Adapted Consulting AS (Norway).*',
    'author': 'Adrian Fjellberg',
    'author_email': 'adrian@adapted.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
